"""Package to render images in Rich and Textual.

Provides Rich Renderables and a Textual Widget to render images via the
Terminal Graphics Protocol (<https://sw.kovidgoyal.net/kitty/graphics-protocol/>),
allowing compatible terminals (i.e. Kitty) to display image data.
"""
